/*
 Stephan Kottler
 kottlers@informatik.uni-tuebingen.de
 *
 */

#include "BasicSolver.hpp"
#include <algorithm>



BasicSolver::BasicSolver(Settings& sets, Statistics& stats):
	ConstraintDB				(sets, stats),
	var_act_hp					( sets.var_max_act , sets.var_min_inc, sets.var_decay),
	state_ok					( true  ),
	queue_begin					( 0	),
	simplify_assign_idx			( 0	),
	simplify_props_remaining	( 0	),
	clause_stamp				( 0 ),
	curr_max_conflicts			( sets.conflict_max_init),
	curr_max_learnts			( 0	)
{ }

BasicSolver::~BasicSolver(){}

//main function
TrBool BasicSolver::isSAT(  double timeout){

	var_flags.ensureSize(numberVariables());

	if(!state_ok)			return FALSE_TrBool;

	curr_max_learnts = numberAllClauses() * settings.learntsize_factor;

	TrBool  status			= UNKNOWN_TrBool;

	if(settings.console_out != NULL ){
		settings.console_out->writeCenteredToBG("SApperloT ... YOU GOT ME SEARCHING ...");
		settings.console_out->writeCenteredToBG(basename(settings.cnf_file_name));
		statistics.printStatsProgressHead();
		settings.console_out->writeString(
				"|____________________________________________________________________________________________|\n");
	}

	while(timeout < 0 || statistics.cpuUserTime() <= timeout){

		//execute big simplify after each 15 restarts - this is instead of preprocessing but may take quite some time
		if((++statistics.nbr_restarts % 15) == 0 ){
			status = bigSimplify();
			if(status.isKnown()) break;
		}

		//search for a satisfying assignment
		status		= search( );

		if(settings.console_out != NULL){ statistics.printStatsProgress(numberLiterals());}

		if(status.isKnown()) break;

		curr_max_conflicts  *= settings.conflict_max_inc;
		curr_max_learnts	*= settings.learntsize_inc;

		backjumpKeepLevel(0,BACKJUMP_SET_PREV);

		//increment the probability for random decisions
		settings.random_var_freq += 0.001;
	}

	if(status.isUnknown()){
		assert(timeout >= 0 && timeout <= statistics.cpuUserTime());
		if(settings.console_out != NULL)
			settings.console_out->writeString("|!!! Timeout of %i seconds exceeded !!! ",(int)timeout);
			return UNKNOWN_TrBool;
	}

	if(settings.console_out != NULL){
		settings.console_out->writeString("|____________________________________________________________________________________________");
		settings.console_out->writeString("| ");
	}

	assert(!status.isTrue() || debugVerifyModel());

	return status;
}

bool BasicSolver::addClause(vector<Literal>& lits){

	assert(decisionLevel() == 0);

	if (!state_ok || lits.empty()){								return false;}

	// Check if clause is satisfied and remove false/duplicate literals:
	std::sort(lits.begin(),lits.end());
	reserveForVariables(lits.back().toVar().idx()+1);

	Literal						last	(INVALID_LIT);
	uint32_t					i	(0);
	vector<Literal>::iterator   to	(lits.begin());
	while(i < lits.size()){
		if(litIsTrue(lits[i]) || lits[i] == last.flipped()){	return true;}

		if(litIsUnknown(lits[i]) && lits[i] != last){
			*to++ = last = lits[i];
			if(!var_act_hp.contains(last.toVar())){ var_act_hp.lazyAddNewElement(last.toVar(),var_act_t(0));}
		}
		++i;
	}

	return addValidClause(lits.begin(),to);
}



/** Add a clause that does not contain duplicate literals or false literals*/
bool BasicSolver::addValidClause  (vector<Literal>::iterator lbeg, const vector<Literal>::iterator lend){

	assert(decisionLevel() == 0);
	uint32_t l_sz(lend-lbeg);

	if(l_sz < 4){
		if(l_sz == 1){
			assert(litIsUnknown(*lbeg));
			uncheckedEnqueue<Clause*>(*lbeg);						return state_ok = (bcp(false) == BCP_SUCCESS);}

		if(l_sz == 2){clsCreateBinary(*lbeg,*(lbeg+1));				return true;}

		if(l_sz == 3){clsCreateTernary(*lbeg,*(lbeg+1),*(lbeg+2));	return true;}

		assert(l_sz == 0);											return state_ok = false;
	}

	assert(l_sz <= Clause::MAX_SIZE);

	if(clsCreate(lbeg,lend) != NULL){								return true;}
	exit(EXIT_FAILURE);
}


TrBool BasicSolver::search( ){

	assert(state_ok);
	uint32_t			max_learnts			= (uint32_t) curr_max_learnts;
	TrBool				result				= UNKNOWN_TrBool;

	uint32_t 			learnts_allowed		= max_learnts;
	int32_t 			conflicts_remain	= curr_max_conflicts < 0 ? (~zero32>>1): (uint32_t) curr_max_conflicts;

	while(conflicts_remain >= 0 || decisionLevel() == 0 ){

		//aply boolean constraint propagation
		int32_t bcp_res	=	bcp(true);

		if(bcp_res == BCP_SUCCESS){		//there is no conflict

			if (decisionLevel() == 0	&& (result = simplify(), result.isKnown())) {				return result;		}

			learnts_allowed = max_learnts +numberAssignments();
			if(learnts_allowed <= numberLearntLongClses()){learntsGC((learnts_allowed * 3)/4); }

			Literal next	= nextDecision();

			if (next == INVALID_LIT)	{															return TRUE_TrBool;   }

			// Increase decision level and enqueue 'next'
			assert(litIsUnknown(next));
			newDecisionLevel();
			uncheckedEnqueue<Clause*>(next);

			//check both polarities at first decision level
			if(decisionLevel() == 1){firstDecisionCheckBothPol(next);}

			set_max_int(statistics.max_dec_level,decisionLevel());


		}else{						//there is a conflict

			++statistics.nbr_conflicts;
			--conflicts_remain;

			if (!firstUIP(bcp_res))  {																return TrBool(state_ok = false); }
			assert(litAssignLevel(conflict_lemma[0]) == decisionLevel());

			//try to minimise the learnt clause
			reduceSelfImpl();

			//determine the level to which to backtrack
			uint32_t backtrack_level = backtrackLevel();
			statistics.storeBackjumpingDistance(decisionLevel() - backtrack_level);

			//finally backtrack
			backjumpKeepLevel(backtrack_level,BACKJUMP_SET_PREV);
			assert(litIsUnknown(conflict_lemma[0]));

			//learn the conflict clause
			const uint32_t l_sz(conflict_lemma.size());
			if(l_sz > 3){
				assert(l_sz <= Clause::MAX_SIZE);
				Clause* learnt_cl = clsCreate(conflict_lemma.begin(),conflict_lemma.end(),true);
				clsBump(learnt_cl);
				uncheckedEnqueue(conflict_lemma[0], learnt_cl);
			}else if(l_sz == 3){
				clsCreateTernary(conflict_lemma[0],conflict_lemma[1],conflict_lemma[2]);
				uncheckedEnqueue(conflict_lemma[0],conflict_lemma[1],conflict_lemma[2]);
			}else if(l_sz == 2){
				clsCreateBinary(conflict_lemma[0],conflict_lemma[1]);
				uncheckedEnqueue(conflict_lemma[0],conflict_lemma[1]);
			}else{
				uncheckedEnqueue<Clause*>(conflict_lemma[0]);
			}

			var_act_hp.decayAll();
			claDecayActivity();

			//panic if formula gets too big
			if ( numberLiterals() > settings.max_literals){
				learntsGC(numberLearntLongClses() / 3);
				++statistics.nbr_panic_gc;
				backjumpKeepLevel(0,BACKJUMP_SET_PREV_FLP);
				break;
			}
		}
	}

	//maximal number of conflicts reached
	return UNKNOWN_TrBool;

}

void	BasicSolver::backjumpUndoAssigns (uint32_t assign_idx,const BackjumpType type){

	Variable x = INVALID_VAR;
	assert(assign_stack.size() >= assign_idx);

	queue_begin 	= assign_stack.size();

	if(type == BACKJUMP_SET_PREV){
		while(queue_begin != assign_idx ){
			x  = assign_stack[--queue_begin].toVar();
			varClearAssignmentSetPrev(x);
			reinsertVarOrder(x);}
	}else if(type == BACKJUMP_SET_PREV_FLP){
		while(queue_begin != assign_idx ){
			x  = assign_stack[--queue_begin].toVar();
			varClearAssignmentSetPrevFlp(x);
			reinsertVarOrder(x);}
	}else{
		while(queue_begin != assign_idx ){
			x  = assign_stack[--queue_begin].toVar();
			varClearAssignmentSimple(x);
			reinsertVarOrder(x);}
	}

	assign_stack.resize(queue_begin);

	//invalidate clause timestamps after backtracking
	if(++clause_stamp > Clause::MAX_STAMP){
		clsClearAllStamps();
		clause_stamp = 0;
		++statistics.nbr_expensive_cls_stamp;
	}
}


Literal BasicSolver::nextDecision(){

	Variable var = INVALID_VAR;

	if(settings.randy.nextDouble() < settings.random_var_freq){
		//random decision
		do{
			if(var_act_hp.empty()) return INVALID_LIT;
			var	= var_act_hp.removeAtPos(settings.randy.nextInt(var_act_hp.size()));

		}while(varValue(var).isKnown());

		++statistics.nbr_rnd_decisions;

	}else{
		//activity based decision
		do{
			if(var_act_hp.empty()) return INVALID_LIT;
			var	= var_act_hp.removeMax();
		}while(varValue(var).isKnown() );

		++statistics.nbr_act_decisions;
	}


	assert(var != INVALID_VAR );

	switch (settings.polarity_mode){
		case Polarity_False	: return var.toNegLit();
		case Polarity_True	: return var.toPosLit();
		case Polarity_Rand	: return settings.randy.nextBool() ?
									var.toNegLit():var.toPosLit();
		case Polarity_Prev	: return varPrevValue(var).isTrue() ?
									var.toPosLit() : var.toNegLit();
		default: assert(false); return var.toNegLit();
	}
}

//for the first decision level check both polarities of a variable
void	BasicSolver::firstDecisionCheckBothPol(Literal next){
	const uint32_t impl_start = assign_stack.size();
	assert(assign_stack[impl_start-1] == next);
	assert(tmp_literal_store.empty());

	//propagate one polarity
	if(BCP_SUCCESS == bcp(false)){
		if(impl_start == assign_stack.size()){return;}

		//keep assignments
		for(uint32_t i = assign_stack.size();i != impl_start;){
			tmp_literal_store.push_back(assign_stack[--i]);};

		backjumpUndoAssigns(impl_start-1,BACKJUMP_PREV_UNTOUCHED);
		uncheckedEnqueue<Clause*>(next.flipped());
		assert(decisionLevel() == 1);

		//propagate opposit polarity
		if(BCP_SUCCESS == bcp(false)){
			uint32_t to = 0;

			for(uint32_t i = 0; i != tmp_literal_store.size();++i){
				Literal lit = tmp_literal_store[i];
				if(litIsTrue(lit)){tmp_literal_store[to++] = lit;}
				else if(litIsFalse(lit)){
					//two literals are equal
					if(!containsBinary(next.flipped(),lit)){clsCreateBinary(next.flipped(),lit);}
					if(!containsBinary(next,lit.flipped())){clsCreateBinary(next,lit.flipped());}
				}
			}

			//enque all unit implications
			if(to != 0){
				backjumpKeepLevel(0,BACKJUMP_PREV_UNTOUCHED);
				statistics.nbr_flipped_prop_learnts += to;
				while(to--){uncheckedEnqueue<Clause*>(tmp_literal_store[to]);}
			}else{
				backjumpUndoAssigns(impl_start-1,BACKJUMP_PREV_UNTOUCHED);
				uncheckedEnqueue<Clause*>(next);
			}

		}else{
			backjumpKeepLevel(0,BACKJUMP_PREV_UNTOUCHED);
			++statistics.nbr_flipped_prop_learnts;
			uncheckedEnqueue<Clause*>(next);
		}
	}else{
		backjumpKeepLevel(0,BACKJUMP_PREV_UNTOUCHED);
		++statistics.nbr_flipped_prop_learnts;
		uncheckedEnqueue<Clause*>(next.flipped());
	}
	tmp_literal_store.clear();
}



//boolean constraint propagation
int32_t	BasicSolver::bcp(const bool init_conflict){

	int32_t			nbr_pivot_lits = BCP_SUCCESS; //indicates no conflict by now
	uint32_t		num_props(0);

	while(queue_begin < assign_stack.size()){

		++num_props;
		nbr_pivot_lits = bcpPropagateLit(init_conflict,assign_stack[queue_begin++].flipped());
	}

	statistics.nbr_propagations	+= num_props;
	simplify_props_remaining		-= num_props;

	return nbr_pivot_lits;
}

//propagate one literal
int32_t	BasicSolver::bcpPropagateLit(const bool init_conflict, const Literal& false_lit){

	we_vec& watched	= litWatchedClauses(false_lit); //watched clauses that have to be checked
	we_iter iter = watched.begin();
	we_iter copy = iter;
	we_iter end  = watched.end();

	while(iter != end){
		WatchedElement& we = *iter++;

		//might be any other literal of this clause except the watched literal
		const Literal other = we.getFirstLiteral();
		assert(other != false_lit);

		//keep this watcher if other lit is true
		if(litIsTrue(other)){ *copy++ = we; continue;}

		if(we.isBinary()){
			*copy++ = we; //keep this watcher
			if(litIsUnknown(other)){
				uncheckedEnqueue<const Literal&>(other,false_lit);
				continue;
			}else{ //other literal in binary clause is false --> conflict
				bcpFinishOnConflict(watched,iter,copy,end);
				return init_conflict ? bcpInitConflict(decisionLevel(),false_lit,other):1;
			}
		}


		if(we.isTernary()){
			*copy++ = we; //keep this watcher
			const Literal third = we.getSecondLiteral();
			if(litIsTrue(third)){
				//change the order of literals to may be realize earlier that this clause is sat
				we.setFirstLiteral(third);
				we.setSecondLiteral(other);
				continue;
			}

			if(litIsUnknown(other)){
				if(litIsFalse(third)){uncheckedEnqueue(other,false_lit,third);}
				continue;
			}else{
				assert(litIsFalse(other));
				if (litIsFalse(third)){
					bcpFinishOnConflict(watched,iter,copy,end);

					if(init_conflict){
						int32_t  nbr_pivot_lits = bcpInitConflict(decisionLevel(),false_lit,other);
						firstUIPCheckLiteral(third,decisionLevel(),nbr_pivot_lits);
						assert(nbr_pivot_lits != 1);
						return nbr_pivot_lits;
					}else{
						return 1; //any pos value
					}
				}
			}
			assert(litIsFalse(other) && litIsUnknown(third));
			uncheckedEnqueue(third,false_lit,other);
			continue;
		}

		//ok we have to load the big clause
		Clause& cl = *(we.getClause());

		uint32_t other_watcher_index = (cl[1] != false_lit);
		assert(cl[other_watcher_index ^ 1] == false_lit && cl[other_watcher_index] != false_lit);

		//it might be that the shortcut literal in 'we' does not point to the actual watched literal anymore
		if(litIsTrue(cl[other_watcher_index])){
			assert(!we.isShort());
			we.setFirstLiteral(cl[other_watcher_index]);
			*copy++ = we; //keep this watcher
			continue;
		}

		if(bcpFindNewWatcher(cl,false_lit,other_watcher_index)){continue;}

		// No new watched found -- clause is unit under assignment:
		we.setFirstLiteral(cl[other_watcher_index]);
		*copy++				= we;

		if (litIsFalse(cl[other_watcher_index])){
			if(cl.isLearnt())clsBump(&cl);

			bcpFinishOnConflict(watched,iter,copy,end);

			if(init_conflict){
				const int32_t dec_level = decisionLevel();
				int32_t  nbr_pivot_lits = bcpInitConflict(dec_level,false_lit,cl[other_watcher_index]);
				for(uint32_t k = 2;k<cl.size();++k){firstUIPCheckLiteral(cl[k],dec_level,nbr_pivot_lits);}
				assert(nbr_pivot_lits != 1);
				return nbr_pivot_lits;
			}else{
				return 1; //any positive value
			}

		}else{
			uncheckedEnqueue(cl[other_watcher_index], &cl);
		}

	}
	if(copy != end) watched.erase(copy,end);

	return BCP_SUCCESS;
}



/**Analyse the conflict
 * nbr_pivot_elements states how many literals of the current decision level are marked*/
bool BasicSolver::firstUIP(int32_t nbr_pivot_elements){
	assert(nbr_pivot_elements >= 0);
	if(nbr_pivot_elements == 0) return false; // learned the empty clause

	const uint32_t   dec_level		(decisionLevel());
	uint32_t		assign_stack_index  = assign_stack.size() ;

	//either none or at least two literals have to be marked, otherwise unit propagation is not implemented properly
	assert(nbr_pivot_elements != 1);
	AssignInfo 	confl_info;
	Literal 	pivot_lit;

	while(true){

		//skip all unseen assignments
		//since there always exists a UIP the index cannot become negative
		while (!var_flags.isSet(assign_stack[--assign_stack_index].toVar())){}
		assert(assign_stack_index >= 0);

		pivot_lit					= assign_stack[assign_stack_index];
		//can be already cleared since this lit cannot appear in any remaining (previous) reason because literals
		//in reason clauses are all false (except for the one it is the assign reason)
		var_flags					.clear(pivot_lit.toVar());

		if(--nbr_pivot_elements == 0){
			assert(litAssignLevel(pivot_lit) == dec_level);
			assert(pivot_lit != INVALID_LIT);
			conflict_lemma[0]				= pivot_lit.flipped();
			return true;
		}

		confl_info					= varAssignInfo(pivot_lit.toVar());
		assert(litAssignLevel(pivot_lit) == dec_level);

		assert(!confl_info.noReason());

		if(confl_info.isShortReason()){
			firstUIPCheckLiteral(confl_info.getShortFirstReason(),dec_level,nbr_pivot_elements);

			if(confl_info.isTernReason()){
				firstUIPCheckLiteral(confl_info.getShortSecondReason(),dec_level,nbr_pivot_elements);
			}

		}else{
			assert(confl_info.isClsReason() && confl_info.getClsReason() != NULL);

			Clause& cl					= *confl_info.getClsReason();
			const uint32_t cl_sz			(cl.size());

			assert(pivot_lit == cl[0]|| pivot_lit == cl[1]);
			assert(litAssignLevel(cl[0]) == dec_level && litAssignLevel(cl[1]) == dec_level);

			if(cl.isLearnt())				clsBump(&cl);

			uint32_t cl_index = pivot_lit == cl[0];//pivot_lit ==cl[0] ? 1 : 0
			assert(cl[cl_index ^1] == pivot_lit);

			//do for all except the pivot_lit
			firstUIPCheckLiteral(cl[cl_index],dec_level,nbr_pivot_elements);
			for(cl_index = 2;cl_index < cl_sz;++cl_index){
				firstUIPCheckLiteral(cl[cl_index],dec_level,nbr_pivot_elements); }
		}
	}
	return false;
}

uint32_t BasicSolver::backtrackLevel(){
	assert(litAssignLevel(conflict_lemma[0]) == decisionLevel());
	uint32_t lsz = conflict_lemma.size();
	if(lsz < 2) return 0;

	uint32_t max_idx = 1;
	while(--lsz > 1){
		if(litAssignLevel(conflict_lemma[lsz]) > litAssignLevel(conflict_lemma[max_idx])){max_idx = lsz;}}

	//such that the literals with maximum level are the watchers
	//this is important for backtracking!
	Literal highest				= conflict_lemma[max_idx];
	conflict_lemma[max_idx]	= conflict_lemma[1];
	conflict_lemma[1]		= highest;


	return litAssignLevel(conflict_lemma[1]);
}



//try to minimise the conflict clause
void	BasicSolver::reduceSelfImpl(){

	if(conflict_lemma.size() < 2){assert(var_flags.allCleared()); return;}

	//sort literals in the clause acording to activity values -> improves propagation in later search process
	sort(conflict_lemma.begin()+1,conflict_lemma.end(),LitActDescendingFunctor(this));
	assert(var_flags_to_clear.empty());

	uint cp,idx;
	for(cp = idx = 1; idx != conflict_lemma.size(); ++idx){
		if(litIsNecessary(conflict_lemma[idx])) {conflict_lemma[cp++] = conflict_lemma[idx];}}

	conflict_lemma.resize(cp);

	unsetConfLemmaFlags();
}

/**Checks if the given literal is needed in the conflict clause.
 * If it can be resolved from the current reasons it can be omitted*/
bool	BasicSolver::litIsNecessary(Literal lit){

	if(varAssignInfo(lit.toVar()).noReason()) return true;

	const Variable start(lit.toVar());

	uint32_t unmark_top(var_flags_to_clear.size());
	assert(tmp_literal_store.empty());

	while(true){

		assert(!varAssignInfo(lit.toVar()).noReason());
		const AssignInfo& a_info = varAssignInfo(lit.toVar());

		bool is_nec;

		if(a_info.isShortReason()){
			is_nec = reduceSelfImpCheckLiteral(a_info.getShortFirstReason());

			if(!is_nec && a_info.isTernReason()){
				is_nec = reduceSelfImpCheckLiteral(a_info.getShortSecondReason());
			}
		}else{
			assert(a_info.isClsReason() && a_info.getClsReason() != NULL);

			Clause& cl	= *(a_info.getClsReason());

			uint32_t cl_index = lit.toVar() == cl[0].toVar();
			assert(cl[cl_index^1].toVar() == lit.toVar());
			assert(var_flags.isSet(cl[cl_index^1].toVar()));

			is_nec = reduceSelfImpCheckLiteral(cl[cl_index]);
			for(cl_index = 2;!is_nec && cl_index < cl.size();++cl_index){
				is_nec = reduceSelfImpCheckLiteral(cl[cl_index]);
			}
		}

		if(is_nec){
			for( ;unmark_top < var_flags_to_clear.size();++unmark_top)   {var_flags.clear(var_flags_to_clear[unmark_top]);}
			for(int i = tmp_literal_store.size();i--;)				{var_flags.clear(tmp_literal_store[i].toVar());}
			tmp_literal_store.clear();
			return true;
		}


		if(tmp_literal_store.empty()){
			statistics.total_lemma_shortening += 1;
			var_flags_to_clear.push_back(start);
			return false;
		}

		lit = tmp_literal_store.back(); tmp_literal_store.pop_back();
		var_flags_to_clear.push_back(lit.toVar());
	}
}


/**Remove all satisfied clauses*/
TrBool BasicSolver::simplify( ){

	assert(decisionLevel() == 0);

	if (!state_ok || bcp(false) != BCP_SUCCESS){   return TrBool(state_ok = false);}

	if (assign_stack.size() != simplify_assign_idx  && (simplify_props_remaining <= 0)){

		++statistics.nbr_simplify_execs;

		removeSat(learnt_cls);
		removeSat(fix_cls);

		simplifyAllShortWatchers();

		simplify_props_remaining   = numberLiterals();
	}

	return UNKNOWN_TrBool;
}




/**Since short clauses are not stored explicitely but directly inside the watchers lits
 * they have to be deleted directly from there*/
void	BasicSolver::simplifyAllShortWatchers (){
	assert(var_flags.allCleared());
	assert(decisionLevel() == 0);

	//look at all learnt unit clauses since the last call to simplify
	while(simplify_assign_idx < assign_stack.size()){
		Literal lit = assign_stack[simplify_assign_idx++];
		simplifyShortWatchersOf(litWatchedClauses(lit));
		simplifyShortWatchersOf(litWatchedClauses(lit.flipped()));
	}


	for(uint32_t i = var_flags_to_clear.size();i--;){var_flags.carelessClear(var_flags_to_clear[i]);}
	var_flags_to_clear.clear();
	assert(var_flags.allCleared());
}

//look at all watchers of one literal
void	BasicSolver::simplifyShortWatchersOf (we_vec& wes){
	we_iter from,to;

	for(from = to = wes.begin(); from != wes.end();++from){
		if(from->isShort()){
			clearShortWatchers(from->getFirstLiteral().toVar());
			if(from->isTernary()){clearShortWatchers(from->getSecondLiteral().toVar());};
		}else{
			//might be that a variable with fixed value has still other watchers
			//since changes of bigSimplify are not simplified further
			*to++ = *from;
		}
	}
	wes.erase(to,wes.end());
}



/**Do the actual work and check the opposit watchers*/
void BasicSolver::clearShortWatchers(Literal lit){

	assert(litIsUnknown(lit));
	assert(queue_begin == assign_stack.size());

	we_iter from,to,end;
	we_vec& wes = litWatchedClauses(lit);

	from = to = wes.begin(); end = wes.end();
	while(from != end){
	WatchedElement& we = *from++;
	if(we.isShort()){

			if(	litIsTrue(we.getFirstLiteral()) ||
				(we.isTernary() && litIsTrue(we.getSecondLiteral()))){continue;} //delete watcher

			if(litIsFalse(we.getFirstLiteral())){
				assert(we.isTernary());
				assert(litIsUnknown(we.getSecondLiteral()));
				*to++ = WatchedElement(we.getSecondLiteral());
				continue;}

			if(we.isTernary() && litIsFalse(we.getSecondLiteral())){
				assert(litIsUnknown(we.getFirstLiteral()));
				*to++ = WatchedElement(we.getFirstLiteral());
				continue;}
		}

		*to++ = we;
	}
	if(to != end){wes.erase(to,end);}
}



void BasicSolver::clearShortWatchers(Variable var){
	if(var_flags.isSet(var) || varValue(var).isKnown()) return;
	var_flags.setFlag(var);	var_flags_to_clear.push_back(var);

	clearShortWatchers(var.toNegLit());
	clearShortWatchers(var.toPosLit());
}


/**Apply asynchronous branching to shrink clauses but only for clauses with under average size*/
TrBool BasicSolver::bigSimplify(){


	assert(decisionLevel() == 0);

	if(settings.console_out != NULL ){settings.console_out->writeString("Execute big simplify");}

	TrBool status = UNKNOWN_TrBool;

	if (!state_ok || bcp(false) != BCP_SUCCESS){   return  TrBool(state_ok = false);}

	if (assign_stack.size() != 0){

		++statistics.nbr_simplify_execs;

		uint32_t max_cls_sz = numberLiteralsInLongFixClses() / (numberOrigLongClses()+1);

		if(max_cls_sz > 3 ){
			status = removeSatAndShrinkClauses(fix_cls,max_cls_sz);
			if(status.isFalse()) return  status ;
		}else if(simplify_assign_idx < assign_stack.size()) removeSat(fix_cls);

		learntsGC((uint32_t)curr_max_learnts/2);
		max_cls_sz = numberLiteralsInLongLearntClses() / (numberLearntLongClses()+1);
		if(max_cls_sz > 3 ){
			status = removeSatAndShrinkClauses(learnt_cls,max_cls_sz);
   			if(status.isFalse()) return  status ;
		}else if(simplify_assign_idx < assign_stack.size()) removeSat(learnt_cls);


		simplifyAllShortWatchers();

		simplify_props_remaining   = numberLiterals();
	}

	assert(var_flags.allCleared());

	if(settings.console_out != NULL ){settings.console_out->writeString("Leave big simplify");}

	return UNKNOWN_TrBool;
}


/**Apply asynchronous branching for all clauses in the given vector up to the specified size*/
TrBool 	BasicSolver::removeSatAndShrinkClauses(vector<Clause*>& clses,uint32_t max_cls_sz){

	TrBool status = UNKNOWN_TrBool;

	vector<Clause*>::iterator to, iter;
	for (iter = to = clses.begin(); iter != clses.end(); ++iter){


		Clause& cl = **iter;

		if(cl.size() > max_cls_sz){
			if (clsCanBeOmitted(cl)){
				statistics.nbr_simplify_worked += 1;
				clsUncreate(*iter);
			}else{
				*to++ = *iter;
			}

			continue;
		}

		uint32_t new_cls_sz = clsCheckSatAndShrink(cl);

		//might be that clause was shrinked and has to be stored as short clause now
		if(new_cls_sz <= 3){
			if(new_cls_sz == 3){
				assert(litIsUnknown(cl[0]) && litIsUnknown(cl[1]) && litIsUnknown(cl[2]));
				clsCreateTernary(cl[0],cl[1],cl[2]);
			}else if(new_cls_sz == 2){
				assert(litIsUnknown(cl[0]) && litIsUnknown(cl[1]));
				clsCreateBinary(cl[0],cl[1]);
			}else if(new_cls_sz == 1){

				if(litIsUnknown(cl[0])){
					uncheckedEnqueue<Clause*>(cl[0]);
					state_ok = bcp(false) == BCP_SUCCESS;
				}

				if(!state_ok || litIsFalse(cl[0])){
					while(iter != clses.end()){*to++ =*iter++;}
					status = TrBool(state_ok = false);
					break;
				}
			}

			statistics.nbr_simplify_worked += 1;
			clsUncreate(&cl,false);
		}else{
			if(new_cls_sz < cl.size()){
				clsShrinkToSize(cl,new_cls_sz);
				clsLinkWatched(&cl);
			}
			*to++ = &cl;
		}
	}


	clses.erase(to,clses.end());

	return status;
}





//return 0 means that clause is not needed any longer (may be unlinked)
//return cl.size() means that clause has to remain unchanged
//return else -> clause can be shrinked to this size (clause is unlinked in this case)
uint32_t BasicSolver::clsCheckSatAndShrink(Clause& cl){

	assert(decisionLevel() == 0);

	if(litIsTrue(cl[0])){
		clsUnlinkWatched(&cl);
		return 0;
	}

	const uint32_t q_begin = assign_stack.size();
	assert(q_begin == queue_begin);

	uint32_t new_end = cl.size();
	bool cls_linked = true;

	for(uint32_t i = cl.size(); --i;) {
		Literal lit = cl[i];

		if(litIsFalse(lit)){
			if(cls_linked){clsUnlinkWatched(&cl); cls_linked = false;}
			cl[i]		= cl[--new_end];
			assert(litIsFalse(cl[i]));
			continue;
		}

		if(litIsTrue(lit) || (uncheckedEnqueue<Clause*>(lit.flipped()),bcp(false) != BCP_SUCCESS)){

			if(cls_linked){clsUnlinkWatched(&cl);}

			backjumpUndoAssigns(q_begin,BACKJUMP_PREV_UNTOUCHED);

			if(litIsTrue(lit)){return 0;}

			uint32_t c_end = 0;
			while( i != new_end){cl[c_end++] = cl[i++];}

			statistics.nbr_bigsimplify_shrinked += (cl.size() - c_end);
			return c_end;
		}
	}

	backjumpUndoAssigns(q_begin,BACKJUMP_PREV_UNTOUCHED);

	if(new_end < cl.size()){statistics.nbr_bigsimplify_shrinked += (cl.size() - new_end);}

	return  new_end;
}







const char*  basic_solver_cpp_version(){return "BasicSolver.cpp:	$Revision: 1.52 $";}
