/*
 Stephan Kottler
 kottlers@informatik.uni-tuebingen.de
 *
 */

#ifndef BITVECTOR_HPP_
#define BITVECTOR_HPP_

/**
 * T is the type of elements for which flags can be set.
 * It has to implement the function T.idx() to get an integer representing an element.
 */
template<class T>
class BitVector{

	/**Array to store a flag-value for each element.*/
	uint32_t*   bits;

	/**Number of elements that are hold.*/
	uint32_t	size;

	uint32_t	array_sz;

	//construction, assignment, destruction
private:
	/**No copying allowed*/
	BitVector(const BitVector& flg)					{assert(false);};

	inline static uint32_t	offset(uint32_t index)	{return index & 31 ;}

	inline static uint32_t	off_mask(uint32_t index){return 1 << (index & 31) ;}

	inline static uint32_t	ar_idx(uint32_t index)	{return index >> 5 ;}

	/**No assignment allowed*/
	const BitVector& operator=(const BitVector& flg){assert(false);return flg;}

public:

	inline BitVector();
	inline BitVector(uint32_t nbr_elements);
	inline ~BitVector();

	inline void ensureSize(const uint32_t nbr_elements);
	inline bool isSet(const T& t)	const	{uint32_t index = t.idx(); assert(index < size);return bits[ar_idx(index)] & off_mask(index);}
	inline void setFlag(const T& t)			{uint32_t index = t.idx();assert(index < size);bits[ar_idx(index)] |= off_mask(index);}
	inline bool checkAndSet(const T& t);
	inline void clear(const T& t)			{uint32_t index = t.idx();assert(index < size); bits[ar_idx(index)] &= ~off_mask(index);}
	inline void carelessClear(const T& t)	{assert(t.idx() < size); bits[ar_idx(t.idx())] = 0;}
	inline void clearAll()					{ for(uint32_t i(array_sz);i--;){bits[i] = 0; }};
	inline bool allCleared();
};

//Inline Implementation

template<class T>
inline BitVector<T>::BitVector(): bits(NULL),size(0),array_sz(0){}

template<class T>
inline BitVector<T>::BitVector(uint32_t nbr_elements):
	size(nbr_elements),array_sz(ar_idx(size)+(offset(size)> 0)){
	bits   = (uint32_t*)malloc( array_sz * sizeof(uint32_t));
	memset ( bits, 0, array_sz * sizeof(uint32_t) );
}

template<class T>
inline BitVector<T>::~BitVector(){
	if(bits != NULL){
		free(bits);
		bits		= NULL;
		size		= 0;
		array_sz	= 0;
	}
}

template<class T>
inline void BitVector<T>::ensureSize(const uint32_t nbr_elements){
	if(nbr_elements > size){
		size		= nbr_elements;
		array_sz	= ar_idx(size)+(offset(size) > 0);
		if(bits != NULL) free(bits);
		bits   = (uint32_t*)malloc( array_sz * sizeof(uint32_t));
		memset ( bits, 0, array_sz * sizeof(uint32_t) );
	}
}


template<class T>
inline bool BitVector<T>::checkAndSet(const T& t){
	uint32_t index = t.idx();
	assert(index < size);
	bool before = isSet(index);
	setFlag(index);
	return before;
}


template<class T>
inline bool BitVector<T>::allCleared(){
	for(uint32_t i(array_sz);i--;){
		if(bits[i] != 0) return false;
	}
	return true;
}




#define bitvector_hpp_version	  "BitVector.hpp:	  $Revision: 1.4 $"



#endif /*BITVECTOR_HPP_*/
